import 'dart:async';
import 'dart:convert';

import 'package:firebase_messaging/firebase_messaging.dart'
    hide RemoteNotification;
import 'package:flutter_local_notifications/flutter_local_notifications.dart';
import 'package:wyatt_cloud_messaging_bloc_base/wyatt_cloud_messaging_bloc.dart';

typedef CloudMessagingStreamController = StreamController<RemoteNotification>;

class CloudMessagingFirebaseDataSource extends CloudMessagingRemoteDataSource {
  final FirebaseMessaging _firebaseMessagingInstance =
      FirebaseMessaging.instance;
  final FlutterLocalNotificationsPlugin _localNotificationInstance =
      FlutterLocalNotificationsPlugin();

  @override
  Future<Stream<RemoteNotification>> onNotification() {
    throw UnimplementedError();
  }

  @override
  Future<Stream<RemoteNotification>> onNotificationBackground() async {
    final backgroundStreamController = CloudMessagingStreamController();

    Future<void> getLastBackgroundNotification() async {
      final lastNotification =
          await _firebaseMessagingInstance.getInitialMessage();
      if (lastNotification != null) {
        backgroundStreamController.add(
          RemoteNotification(
            title: lastNotification.notification?.title,
            body: lastNotification.notification?.body,
            data: lastNotification.data,
          ),
        );
      }

      FirebaseMessaging.onMessageOpenedApp.listen((notification) {
        backgroundStreamController.add(
          RemoteNotification(
            title: notification.notification?.title,
            body: notification.notification?.body,
            data: notification.data,
          ),
        );
      });
    }

    unawaited(getLastBackgroundNotification());

    return backgroundStreamController.stream;
  }

  @override
  Future<Stream<RemoteNotification>> onNotificationForeground() async {
    await _firebaseMessagingInstance
        .setForegroundNotificationPresentationOptions(
      alert: true,
      badge: true,
      sound: true,
    );

    const AndroidNotificationChannel channel = AndroidNotificationChannel(
      'high_importance_channel',
      'High Importance Notifications',
      description: 'This channel is used for important notifications.',
      importance: Importance.max,
    );

    const initializationSettingsAndroid = AndroidInitializationSettings(
      '@drawable/ic_stat_onesignal_default',
    );

    const initializationSettings = InitializationSettings(
      android: initializationSettingsAndroid,
      iOS: DarwinInitializationSettings(),
    );

    final CloudMessagingStreamController foregroundStream =
        CloudMessagingStreamController();

    await _localNotificationInstance.initialize(
      initializationSettings,
      onDidReceiveNotificationResponse: (message) {
        foregroundStream.add(
          RemoteNotification(
            data:
                jsonDecode(message.payload.toString()) as Map<String, dynamic>,
          ),
        );
      },
    );

    await _localNotificationInstance
        .resolvePlatformSpecificImplementation<
            AndroidFlutterLocalNotificationsPlugin>()
        ?.createNotificationChannel(channel);

    FirebaseMessaging.onMessage.listen(
      (message) {
        final notification = RemoteNotification(
          title: message.notification?.title,
          body: message.notification?.body,
        );

        final android = message.notification?.android;

        if (message.notification != null && android != null) {
          _localNotificationInstance.show(
            notification.hashCode,
            notification.title,
            notification.body,
            NotificationDetails(
              android: AndroidNotificationDetails(
                channel.id,
                channel.name,
                channelDescription: channel.description,
                icon: android.smallIcon,
              ),
            ),
            payload: jsonEncode(message.data),
          );
        }
      },
    );

    return foregroundStream.stream;
  }

  @override
  Future<String?> getToken() => _firebaseMessagingInstance.getToken();

  @override
  Future<void> requestPermissions() =>
      _firebaseMessagingInstance.requestPermission();

  @override
  Future<void> subscribeToTopic(String topic) =>
      _firebaseMessagingInstance.subscribeToTopic(topic);

  @override
  Future<void> unsubscribeFromTopic(String topic) =>
      _firebaseMessagingInstance.unsubscribeFromTopic(topic);
}
